/*
 * WL24xx Wireless LAN Card Common APIs
 *
 * Design Issues:
 *
 *   1. Prevent using O/S dependent routines as possible.
 *   2. Fully utilize Rx/Tx queueing rings.
 *   3. Avoid additional memcpy between O/S and card if its size is large.
 *      (But if the overhead is little, I prefer to make the program clear.)
 *
 * References:
 *
 *   1. WL24xx packet drivers (tooasm.asm)
 *   2. Access Point Firmware Interface Specification for IEEE 802.11 SUTRO
 *   3. IEEE 802.11
 *
 * Additional Information:
 *
 *   1. To prevent forward declaration, I put primitive functions ahead.
 *   2. All veriables are named with type, with form <type><Name> e.g.:
 *      cTmpVar -> type is 'c', and Name is 'TmpVar'
 *
 *      Type List:
 *          c   - unsigned char (BYTE)
 *          n   - int           (DWORD)
 *          b   - boolean int   (BOOL, DWORD)
 *          u   - unsigned int  (WORD)
 *          ul  - unsigned long (DWORD)
 *          p   - pointer       (DWORD)
 *          psz - pointer to string ended with zero
 *          cpsz- const psz
 *          q   - queue (structure)
 *      Others are defined by Windows programming convensions.
 *
 * To do:
 *
 *   1. Fully compatiable with IEEE 802.11 (SUTRO need to be refined too)
 *   2. Process Broadcast Queue , PsPoll Queue, and CfPoll Queue
 *
 */
#include "./wlapi.h"

WHEN_WLAPI_DEBUG(DWORD WL_DbFlag = 0x00000000L);

static BYTE WL_FPAGE[] = { BSS_FPAGE0, BSS_FPAGE1, BSS_FPAGE2, BSS_FPAGE3 };

static const BYTE WL_SIGNAL_TABLE[2] = { 
    0x0A, 
    0x14   /* QPSK */
};


/*
 * For debugging
 */
int  TraceNothing(const char *cpszFormat, ...)
{
    return 0;
}

void TraceBuf(void *pBuf, int nSize)
{
    int     i;
    BYTE    *pCh = (BYTE *) pBuf;
    
    for (i = 0; i < nSize; i++)
        printk("%2.2x ", pCh[i]);
        
    printk("\n");
}

/*
 * Get Ethernet MAC addresss.
 *
 * WARNING: We switch to FPAGE0 and switc back again.
 *          Making sure there is no other WL function beening called by ISR.
 */
BOOL WL_GetFlashMacAddress(WL_Adaptor *pAdaptor)
{
    int   nBaseAddr = pAdaptor->nBaseAddr;
    
    ASSERT_STAT((pAdaptor->ulDbgStat & WL_STAT_TRIVIAL_FUNC) == 0);
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat |= WL_STAT_FLASH_READ);
    
    /* get MAC Address */
    OutB_P(BSS_FPAGE3, nBaseAddr + NIC_BSS);  /* BSS */
    OutB_P(0x00, nBaseAddr + NIC_LMAL);       /* LMAL */
    OutB_P(0x40, nBaseAddr + NIC_LMAH);       /* LMAH */

    /* wait for reading EEPROM */
    NOPLOOP(100);
    pAdaptor->MacAddress.b0 = InB(nBaseAddr + NIC_IODPA);
    NOPLOOP(100);
    pAdaptor->MacAddress.b1 = InB(nBaseAddr + NIC_IODPA);
    NOPLOOP(100);
    pAdaptor->MacAddress.b2 = InB(nBaseAddr + NIC_IODPA);
    NOPLOOP(100);
    pAdaptor->MacAddress.b3 = InB(nBaseAddr + NIC_IODPA);
    NOPLOOP(100);
    pAdaptor->MacAddress.b4 = InB(nBaseAddr + NIC_IODPA);
    NOPLOOP(100);
    pAdaptor->MacAddress.b5 = InB(nBaseAddr + NIC_IODPA);
    NOPLOOP(100);
    pAdaptor->FrequencyDomain = InB(nBaseAddr + NIC_IODPA);
    NOPLOOP(100);
    OutB_P(BSS_FPAGE0, nBaseAddr + NIC_BSS); 
    OutB_P(0x04, nBaseAddr + NIC_LMAL);
    OutB_P(0x40, nBaseAddr + NIC_LMAH);
    NOPLOOP(100);
    pAdaptor->Version[0] = InB(nBaseAddr + NIC_IODPA);
    NOPLOOP(100);
    pAdaptor->Version[1] = InB(nBaseAddr + NIC_IODPA);
    /* switch to SRAM Page 0 (for safety) */
    OutB(BSS_SPAGE0, pAdaptor->nBaseAddr + NIC_BSS);
    
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_FLASH_READ);

    /* Z-Com's MAC Address should be 00:60:... */
    return pAdaptor->MacAddress.b0 == 0x00 && pAdaptor->MacAddress.b1 == 0x60;
}

void WL_FlashOutB(WL_Adaptor *pAdaptor, WORD uPage, WORD uAddr, BYTE cData)
{
    /* switch to Flash RAM Page uPage */
    OutB(WL_FPAGE[uPage], pAdaptor->nBaseAddr + NIC_BSS);
    
    /* set LMAL and LMAH */
    OutB(uAddr & 0xff, pAdaptor->nBaseAddr + NIC_LMAL);
    OutB((uAddr >> 8) & 0x7f,   pAdaptor->nBaseAddr + NIC_LMAH);
    
    /* out data to Port A */
    OutB(cData, pAdaptor->nBaseAddr + NIC_IODPA);
}

void WL_FlashOutB128(WL_Adaptor *pAdaptor, WORD uPage, WORD uAddr, char *pbuf)
{
    int i;
    /* switch to Flash RAM Page uPage */
    OutB(WL_FPAGE[uPage], pAdaptor->nBaseAddr + NIC_BSS);
    
    /* set LMAL and LMAH */
    OutB(uAddr & 0xff, pAdaptor->nBaseAddr + NIC_LMAL);
    OutB((uAddr >> 8) & 0x7f, pAdaptor->nBaseAddr + NIC_LMAH);
    
    /* out data to Port A */
    for (i=0;i<128;i++){
       OutB(*pbuf, pAdaptor->nBaseAddr + NIC_IODPA);
       pbuf++;
    }
    NOPLOOP(10000);
}

BYTE WL_FlashInB(WL_Adaptor *pAdaptor, WORD uPage, WORD uAddr)
{
    /* switch to Flash RAM Page uPage */
    OutB(WL_FPAGE[uPage], pAdaptor->nBaseAddr + NIC_BSS);
    
    /* set LMAL and LMAH */
    OutB(uAddr & 0xff, pAdaptor->nBaseAddr + NIC_LMAL);
    OutB(uAddr >> 8,   pAdaptor->nBaseAddr + NIC_LMAH);
    
    /* out data to Port A */
    return InB(pAdaptor->nBaseAddr + NIC_IODPA);
}

void WL_FlashInSB(WL_Adaptor *pAdaptor, WORD uPage, WORD uSrc, void *pDest, int nSize)
{
    /* switch to SRAM Page 0 */
    OutB(WL_FPAGE[uPage], pAdaptor->nBaseAddr + NIC_BSS);
    
    /* set LMAL and LMAH */
    OutB(uSrc & 0xff, pAdaptor->nBaseAddr + NIC_LMAL);
    OutB(uSrc >> 8,   pAdaptor->nBaseAddr + NIC_LMAH);
    
    /* rep get from Port A */
    insb(pAdaptor->nBaseAddr + NIC_IODPA, pDest, nSize);
}

/*
 * When calling this function, must hold SUTRO first.
 */
WORD WL_GetFlashID(WL_Adaptor *pAdaptor)
{

    BYTE cByte0, cByte1;
    WORD uID;
    
    /* Autoselect command */
    WL_FlashOutB(pAdaptor, 0, 0x5555, 0xAA);
    WL_FlashOutB(pAdaptor, 0, 0x2AAA, 0x55);
    WL_FlashOutB(pAdaptor, 0, 0x5555, 0x90);
    NOPLOOP(10000);
    
    cByte0 = WL_FlashInB(pAdaptor, 0, 0x0);
    cByte1 = WL_FlashInB(pAdaptor, 0, 0x1);
    
    uID = (cByte0 << 8) | cByte1;
    
    WL_FlashOutB(pAdaptor, 0, 0x5555, 0xAA);
    WL_FlashOutB(pAdaptor, 0, 0x2AAA, 0x55);
    WL_FlashOutB(pAdaptor, 0, 0x5555, 0x0f0);
    NOPLOOP(10000);
/*    printk("Flash ROM ID = 0x%x\n", uID);  */
    return uID;
}

/*
 * Polling if Erase/Programming command is completed
 * Note: IF a==b THEN XOR(a,b) = 0 
 *
 * When calling this function, must hold SUTRO first.
 */
BOOL WL_FlashWriteOk(WL_Adaptor *pAdaptor)
{
    BYTE cByte0, cByte1;

    /* Check 'Toggle Bit' (DQ6) to see if completed */
    do {
        cByte0 = WL_FlashInB(pAdaptor, 0, 0x0);
        cByte1 = WL_FlashInB(pAdaptor, 0, 0x0);
        
        /* Test if exceeded Time Limits (DQ5) */
        if (cByte1 & 0x20) {
            /* Must test DQ6 again before return FALSE */
            cByte0 = WL_FlashInB(pAdaptor, 0, 0x0);
            cByte1 = WL_FlashInB(pAdaptor, 0, 0x0);
    
            return ((cByte0 ^ cByte1) & 0x40) == 0;
        }
            
    } while ((cByte0 ^ cByte1) & 0x40);
    
    return TRUE;
}

/*
 * When calling this function, must hold SUTRO first.
 */
BOOL WL_FlashEraseSector(WL_Adaptor *pAdaptor, WORD uSector)
{
    WORD uPage, uAddr;

    uPage = uSector / 2;
    uAddr = (uSector & 1) ? 0x4000 : 0x0;
    
    /* Sector Erase command (6 commands must within 100 uS) */
    WL_FlashOutB(pAdaptor, 0, 0x5555, 0xAA);
    WL_FlashOutB(pAdaptor, 0, 0x2AAA, 0x55);
    WL_FlashOutB(pAdaptor, 0, 0x5555, 0x80);
    WL_FlashOutB(pAdaptor, 0, 0x5555, 0xAA);
    WL_FlashOutB(pAdaptor, 0, 0x2AAA, 0x55);
    WL_FlashOutB(pAdaptor, uPage, uAddr, 0x30);

    return WL_FlashWriteOk(pAdaptor);
}

/*
 * When calling this function, must hold SUTRO first.
 */
BOOL WL_FlashWriteByte(WL_Adaptor *pAdaptor, WORD uPage, WORD uAddr, BYTE cData)
{
    /* Autoselect command */
    WL_FlashOutB(pAdaptor, 0, 0x5555, 0xAA);
    WL_FlashOutB(pAdaptor, 0, 0x2AAA, 0x55);
    WL_FlashOutB(pAdaptor, 0, 0x5555, 0xA0);

    WL_FlashOutB(pAdaptor, uPage, uAddr, cData);

    return WL_FlashWriteOk(pAdaptor);
}

/*
 * When calling this function, must hold SUTRO first.
 */
BOOL WL_FlashWriteByte128(WL_Adaptor *pAdaptor, WORD uPage, WORD uAddr,char *pbuf)
{
    /* Autoselect command */
    WL_FlashOutB(pAdaptor, 0, 0x5555, 0xAA);
    WL_FlashOutB(pAdaptor, 0, 0x2AAA, 0x55);
    WL_FlashOutB(pAdaptor, 0, 0x5555, 0xA0);

    WL_FlashOutB128(pAdaptor, uPage, uAddr, pbuf);
    return TRUE;
}
/*
 * Write pAdaptor->mibExtra into Flash ROM
 *
 * In fact, only first 29 bytes are used. (not all of extra MIB)
 * To prevent alter other data will be used by future versions, we preserve
 * 256 bytes.
 */
BOOL WL_WriteFlash(WL_Adaptor *pAdaptor,unsigned char *pbuf, WORD uLen)
{
    int  i;
    WORD uFlashID;
    //BYTE buf[256];
    BOOL bRunning;
    DWORD BufAddress;
    bRunning = WL_HoldSutro(pAdaptor);
    
    uFlashID = WL_GetFlashID(pAdaptor);
    
    BufAddress= *(DWORD *) pbuf;
    pbuf+=4;uLen-=4;
    if ((uFlashID == 0x0120) || (uFlashID == 0x016E)){
        /* It's AMD AM29F010, must be erase before programming */
        /* Erase 1st 16 Kbytes within Page 0 */
        if ((BufAddress & 0x3fff) ==0){
            if (!WL_FlashEraseSector(pAdaptor, BufAddress>>14 )) {
                printk("wlapi.c: WL_FlashEraseSector(0) failed!\n");
                if (bRunning)
                    WL_UnHoldSutro(pAdaptor);
                return FALSE;
            }
        }
        for (i = 0; i < uLen; i++) {
           if (!WL_FlashWriteByte(pAdaptor, BufAddress>>15, BufAddress+i, *pbuf)) {
               printk("wlapi.c: WL_FlashWriteByte(buf[%d]) failed!\n", i);
               if (bRunning)
                   WL_UnHoldSutro(pAdaptor);
               return FALSE;
           }
           pbuf++;
       }
       return TRUE;
    }
    else if (uFlashID == 0x1fd5){
        /* It's AT29C010A */
       /* Programming flash ROM byte by byte */
       for (i = 0; i < uLen; i+=128) {
           if (!WL_FlashWriteByte128(pAdaptor, BufAddress>>15, (BufAddress & 0x7fff), pbuf)) {
               printk("wlapi.c: WL_FlashWriteByte(buf[%d]) failed!\n", i);
               if (bRunning)
                   WL_UnHoldSutro(pAdaptor);
               return FALSE;
           }
           pbuf+=128;
           BufAddress+=128;
       }
       return TRUE;
    }
    else {
        /* It's not AT29C010A */
        printk("wlapi.c: Flash ROM type (0x%x) is unknown!\n", uFlashID);
        if (bRunning)
            WL_UnHoldSutro(pAdaptor);
        return FALSE;
    }
    /* Reset command */
/*    WL_FlashOutB(pAdaptor, 0, 0x5555, 0xAA); */
/*    WL_FlashOutB(pAdaptor, 0, 0x2AAA, 0x55); */
/*    WL_FlashOutB(pAdaptor, 0, 0x5555, 0xF0); */
/*    NOPLOOP(10000); */
    
/*    if (bRunning)    */
/*        WL_UnHoldSutro(pAdaptor); */
        
    return TRUE;
}

/*
 * Mask interrupt from SUTRO. (i.e. SUTRO cannot interrupt the HOST)
 * Return: TRUE if interrupt is originally enabled
 */
BOOL WL_BlockInterrupt(WL_Adaptor *pAdaptor)
{
    BYTE    cOld, cNew;
    
    /* ASSERT_STAT((pAdaptor->ulDbgStat & WL_STAT_TRIVIAL_FUNC) == 0); */
    /* WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat |= WL_STAT_BLK_INTR); */
    
    cOld = InB(pAdaptor->nBaseAddr + NIC_GCR);
    cNew = cOld & (~(GCR_ECINT | GCR_INT2EC | GCR_ENECINT));
    OutB(cNew, pAdaptor->nBaseAddr + NIC_GCR);
    
    /* WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_BLK_INTR); */
    return (cOld & GCR_ENECINT);
}

/*
 * Enable interrupt from SUTRO. (i.e. SUTRO can interrupt the HOST)
 * Return: TRUE if interrupt is originally enabled
 */
BOOL WL_UnBlockInterrupt(WL_Adaptor *pAdaptor)
{
    BYTE    cOld, cNew;
    
    ASSERT_STAT((pAdaptor->ulDbgStat & WL_STAT_TRIVIAL_FUNC) == 0);
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat |= WL_STAT_UNBLK_INTR);
    
    cOld = InB(pAdaptor->nBaseAddr + NIC_GCR);
    cNew = (cOld  & ~(GCR_ECINT | GCR_INT2EC)) | GCR_ENECINT;
    OutB(cNew, pAdaptor->nBaseAddr + NIC_GCR);
    
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_UNBLK_INTR);
    return (cOld & GCR_ENECINT);
}

/*
 * Hold SUTRO. (i.e. make SUTRO stop)
 * Return: TRUE if SUTRO is originally running
 */
BOOL WL_HoldSutro(WL_Adaptor *pAdaptor)
{
    BYTE    cOld, cNew;
    
    cOld = InB(pAdaptor->nBaseAddr + NIC_GCR);
    cNew = (cOld & ~(GCR_ECINT | GCR_INT2EC)) | GCR_ECWAIT;
    OutB(cNew, pAdaptor->nBaseAddr + NIC_GCR);

    return (cOld & GCR_ECWAIT) == 0;
}

/*
 * UnHold SUTRO. (i.e. make SUTRO running)
 * Return: TRUE if SUTRO is originally running
 */
BOOL WL_UnHoldSutro(WL_Adaptor *pAdaptor)
{
    BYTE    cOld, cNew;
    
    cOld = InB(pAdaptor->nBaseAddr + NIC_GCR);
    cNew = cOld  & (~(GCR_ECINT | GCR_INT2EC  | GCR_ECWAIT));
    OutB(cNew, pAdaptor->nBaseAddr + NIC_GCR);

    return (cOld & GCR_ECWAIT) == 0;
}

/*
 * Info card to execute a command
 */
void WL_InteruptWla(WL_Adaptor *pAdaptor)
{
    BYTE    cTmp;
    
    ASSERT_STAT((pAdaptor->ulDbgStat & WL_STAT_TRIVIAL_FUNC) == 0);
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat |= WL_STAT_INT_WLA);
    
    cTmp = InB(pAdaptor->nBaseAddr + NIC_GCR);
    cTmp = (cTmp | GCR_INT2EC) &  (~GCR_ECINT);
    OutB(cTmp, pAdaptor->nBaseAddr + NIC_GCR);

    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_INT_WLA);
}

/*
 *  Move 'nSize' bytes from PC to card. (Shouldn't be interrupted)
 *
 *  uDest = Card addressing space
 *  pSrc  = PC addressing space
 *  nSize = Bytes to move
 */
void WL_SetToWla(WL_Adaptor *pAdaptor, WORD uDest, void *pSrc, int nSize)
{
    ASSERT_STAT((pAdaptor->ulDbgStat & WL_STAT_TRIVIAL_FUNC) == 0);
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat |= WL_STAT_SET_WLA);
    
    /* switch to SRAM Page 0 */
    if (uDest & 0x8000){
        OutB(BSS_SPAGE1, pAdaptor->nBaseAddr + NIC_BSS);
    }
    else{
        OutB(BSS_SPAGE0, pAdaptor->nBaseAddr + NIC_BSS);
    }
//    OutB(((uDest>> 11) & 0x18), pAdaptor->nBaseAddr + NIC_BSS);
    
    /* set LMAL and LMAH */
    OutB(uDest & 0xff, pAdaptor->nBaseAddr + NIC_LMAL);
    OutB(((uDest >> 8) & 0x7f),   pAdaptor->nBaseAddr + NIC_LMAH);
    
    /* rep out to Port A */
    OutSB(pAdaptor->nBaseAddr + NIC_IODPA, pSrc, nSize);

    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_SET_WLA);
}

/*
 *  Move 'nSize' bytes from card to PC. (Shouldn't be interrupted)
 *
 *  pSrc  = Card addressing space
 *  uDest = PC addressing space
 *  nSize = Bytes to move
 */
void WL_GetFromWla(WL_Adaptor *pAdaptor, WORD uSrc, void *pDest, int nSize)
{
    ASSERT_STAT((pAdaptor->ulDbgStat & WL_STAT_TRIVIAL_FUNC) == 0);
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat |= WL_STAT_GET_WLA);
    
    /* switch to SRAM Page 0 */
    if (uSrc & 0x8000){
        OutB(BSS_SPAGE1, pAdaptor->nBaseAddr + NIC_BSS);
    }
    else{
        OutB(BSS_SPAGE0, pAdaptor->nBaseAddr + NIC_BSS);
    }
//    OutB(((uSrc>> 11) & 0x18), pAdaptor->nBaseAddr + NIC_BSS);
    
    /* set LMAL and LMAH */
    OutB(uSrc & 0xff, pAdaptor->nBaseAddr + NIC_LMAL);
    OutB(((uSrc >> 8) & 0x7f),   pAdaptor->nBaseAddr + NIC_LMAH);
    
    /* rep get from Port A */
    insb(pAdaptor->nBaseAddr + NIC_IODPA, pDest, nSize);

    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_GET_WLA);
}

/*
 * Get/Allocate a free Tx Data Buffer
 *
 *  *--------------*-----------------*----------------------------------*
 *  |    PLCP      |    MAC Header   |  DST  SRC         Data ...       |
 *  |  (24 bytes)  |    (30 bytes)   |  (6)  (6)  (Ethernet Row Data)   |
 *  *--------------*-----------------*----------------------------------*
 *  \               \- IEEE 802.11 -/ \------------ uDataLen -----------/
 *    \-------- TX_HEADER --------/    \-------- Ethernet Frame -------/
 *
 * Return = Postion in Card
 */
WORD WL_GetTxBuffer(WL_Adaptor *pAdaptor, WORD uDataLen)
{
    WORD  uRet,uNext,uBlockCount,uZero;
    WORD  uFullLen = sizeof(TX_HEADER) + uDataLen;
    if ( uFullLen > pAdaptor->uTxBufferCount*254 ) return 0;
    uRet= pAdaptor->uTxBufferHead;
    uBlockCount=uZero=0;
    while (uFullLen>0){
        if (uFullLen<254) uFullLen=0;
        else uFullLen-=254;
        WL_GetFromWla(pAdaptor, pAdaptor->uTxBufferHead, &uNext, sizeof(uNext));
        if (uFullLen==0){
            WL_SetToWla(pAdaptor, pAdaptor->uTxBufferHead, &uZero, sizeof(uZero));
        }
        pAdaptor->uTxBufferHead=uNext;
        uBlockCount++;
        /* if buffer is not enough */
        if ((uNext==0) && (uFullLen !=0)){
            pAdaptor->uTxBufferHead=uRet;
            return 0;
        }
    }
    pAdaptor->uTxBufferCount-=uBlockCount;
    return uRet;
}

/*
 * Free an allocated Tx Buffer. uPtr must be correct position.
 *
 */
void WL_FreeTxBuffer(WL_Adaptor *pAdaptor, WORD uPtr)
{
    WORD uNext;

    /* check if all space is not free */
    if (pAdaptor->uTxBufferHead == 0){
        pAdaptor->uTxBufferHead = uPtr;
    }
    else{
        WL_SetToWla(pAdaptor, pAdaptor->uTxBufferTail, &uPtr, sizeof(uPtr));
    }
    while(uPtr != 0){
        pAdaptor->uTxBufferCount++;
        WL_GetFromWla(pAdaptor, uPtr, &uNext, sizeof(uNext));
        pAdaptor->uTxBufferTail = uPtr;
        uPtr=uNext;
    }
    return;
}


void WL_GetMibValue(WL_Adaptor *pAdaptor, WORD cIndex)
{
    WORD uPtr;
    struct t_GetReq pGetReq;
    ASSERT_STAT((pAdaptor->ulDbgStat & WL_STAT_MIB_FUNC) == 0);
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat |= WL_STAT_GET_MIB);
    pGetReq.NextBlock=0;
    pGetReq.SignalID=GetRequest;
    pGetReq.MibAtrib=cIndex;
    if (WL_ESBQRequestTest(pAdaptor)){
        uPtr= WL_GetTxBuffer(pAdaptor, sizeof(pGetReq));
        if (uPtr !=0){
            WL_SetToWla(pAdaptor, uPtr, &pGetReq, sizeof(pGetReq));
            WL_ESBQRequest(pAdaptor,&uPtr);
        }
    }
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_GET_MIB);
    return;
}

void WL_SetMibValue(WL_Adaptor *pAdaptor, WORD cIndex, void *pSrc, BYTE cSize)
{

    WORD uPtr;
    struct t_SetReq pSetReq;
    ASSERT_STAT((pAdaptor->ulDbgStat & WL_STAT_MIB_FUNC) == 0);
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat |= WL_STAT_SET_MIB);
    pSetReq.NextBlock=0;
    pSetReq.SignalID=SetRequest;
    pSetReq.MibAtrib=cIndex;
    memcpy((char *) &pSetReq.MibValue,(char *) pSrc,cSize);
    if (WL_ESBQRequestTest(pAdaptor)){
        uPtr= WL_GetTxBuffer(pAdaptor, sizeof(pSetReq));
        if (uPtr !=0){
            WL_SetToWla(pAdaptor, uPtr, &pSetReq, sizeof(pSetReq));
            WL_ESBQRequest(pAdaptor,&uPtr);
        }
    }
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_SET_MIB);
    return;
}

BOOL WL_ResetBoard(WL_Adaptor *pAdaptor)
{
    BYTE    cTmp;
    DWORD   i;
    
    /* Coreset */
    OutB_P(GCR_CORESET, pAdaptor->nBaseAddr + NIC_GCR);
    NOPLOOP(500);
    OutB_P(0,           pAdaptor->nBaseAddr + NIC_GCR);
    NOPLOOP(500);
    OutB_P(GCR_CORESET, pAdaptor->nBaseAddr + NIC_GCR);

    /* Reset SRAM 0x480 to zero */
    cTmp = 0;
    WL_SetToWla(pAdaptor, 0x480, &cTmp, sizeof(cTmp));
    
    /* Start up */
    OutB_P(0, pAdaptor->nBaseAddr + NIC_GCR);
    
    /* NOP */
    NOPLOOP(1024*50);

    /* WL_UnBlockInterrupt(pAdaptor); */

    /* Polling Self_Test_Status */    
    for (i = 0; i < 100; i++) {
        WL_GetFromWla(pAdaptor, 0x480, &cTmp, sizeof(cTmp));

        if (cTmp == 'W') {
            /* firmware complete all test successfully */
            cTmp = 'A';
            WL_SetToWla(pAdaptor, 0x480, &cTmp, sizeof(cTmp));
            return TRUE;
        }
        NOPLOOP(5000);
    }

    return FALSE;
}

BOOL WL_InitFirmware(WL_Adaptor *pAdaptor)
{
    WORD    uPtr,uNext;
    
    ASSERT_STAT((pAdaptor->ulDbgStat & WL_STAT_API_FUNC) == 0);
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat |= WL_STAT_INIT);
    
    if (!WL_ResetBoard(pAdaptor)) {
        TRACE_INIT("wlapi.c: WL_ResetBoard failed!\n");
        WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_INIT);
        return FALSE;
    }
        
    /* Switch to SRAM Page 0 */
    OutB(BSS_SPAGE0, pAdaptor->nBaseAddr + NIC_BSS);
    /* Read parameter from card */
    WL_GetFromWla(pAdaptor, 0x482, &pAdaptor->uESBQReqStart,2);
    WL_GetFromWla(pAdaptor, 0x486, &pAdaptor->uESBQReqEnd,2);
    WL_GetFromWla(pAdaptor, 0x488, &pAdaptor->uESBQConfirmStart,2);
    WL_GetFromWla(pAdaptor, 0x48C, &pAdaptor->uESBQConfirmEnd,2);
    WL_GetFromWla(pAdaptor, 0x48E, &pAdaptor->uTxBufferHead,2);
    WL_GetFromWla(pAdaptor, 0x492, &pAdaptor->uTxBufferSize,2);
    pAdaptor->uESBQReqTail=pAdaptor->uESBQReqHead=pAdaptor->uESBQReqStart;
    pAdaptor->uESBQReqEnd += pAdaptor->uESBQReqStart;
    pAdaptor->uESBQConfirm = pAdaptor->uESBQConfirmStart;
    pAdaptor->uESBQConfirmEnd += pAdaptor->uESBQConfirmStart;
    /* Initial Tx Buffer */
    pAdaptor->uTxBufferCount=1;
    uPtr = pAdaptor->uTxBufferHead;
    uNext = uPtr + BlockSize;
    while ( (uNext- pAdaptor->uTxBufferHead) < pAdaptor->uTxBufferSize){
        pAdaptor->uTxBufferCount++;
        WL_SetToWla(pAdaptor, uPtr, &uNext, sizeof(uNext));
        uPtr = uNext;
        uNext = uPtr + BlockSize;
    }
    uNext = 0 ;
    WL_SetToWla(pAdaptor, uPtr, &uNext, sizeof(uNext));
    pAdaptor->uTxBufferTail=uPtr;
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_INIT);
    return TRUE;    
}

/*
 * Send a packet.
 *
 * pData = Ethernet raw frame.  (e.g. pData[0] - pData[5] is Dest MAC Addr, 
 *                                    pData[6] - pData[11] is Src MAC Addr)
 * Ref: IEEE 802.11
 */
BOOL WL_SendPkt(WL_Adaptor *pAdaptor, BYTE *pData, WORD uLen)
{
    WORD    uBuf,uSig,uNext,templen,packet_length,cTemp;
    BYTE    *DataPtr;
    struct t_MdReq Signal;
    DataPtr=(char *) pData;
    ASSERT_STAT((pAdaptor->ulDbgStat & WL_STAT_API_FUNC) == 0);
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat |= WL_STAT_TX_PKT);

    if (WL_ESBQRequestTest(pAdaptor)){
        uSig = WL_GetTxBuffer(pAdaptor, sizeof(Signal));
        if (uSig == 0) {   /* No free buffer available */
            WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_TX_PKT);
            return FALSE;
        }

        uBuf = WL_GetTxBuffer(pAdaptor, uLen+26+24);

        if (uBuf == 0)  {
            /* No free buffer available */
            WL_FreeTxBuffer(pAdaptor, uSig);
            WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_TX_PKT);
            return FALSE;
        }
        memcpy((char *) &(Signal.DAddress[0]),DataPtr,12);
        packet_length=uLen-12;
        DataPtr +=12;
        Signal.NextBlock=0;
        Signal.SignalID=MdRequest;
        Signal.Data=uBuf;
        if (pAdaptor->LLCType==1){
            if (((*DataPtr)*256+(*(DataPtr+1)))>1500){
                Signal.Size=packet_length+24+4+6;
                cTemp=0xaaaa;
                WL_SetToWla(pAdaptor, uBuf+2+sizeof(TxHeader)-6, &cTemp, sizeof(cTemp));
                cTemp=0x03;
                WL_SetToWla(pAdaptor, uBuf+2+sizeof(TxHeader)-4, &cTemp, sizeof(cTemp));
                cTemp=0x0;
                WL_SetToWla(pAdaptor, uBuf+2+sizeof(TxHeader)-2, &cTemp, sizeof(cTemp));
                if (packet_length>(254-sizeof(TxHeader))){
                    templen=(254-sizeof(TxHeader));
                    packet_length-=templen;
                }
                else{
                    templen=packet_length;
                    packet_length=0;
                }
                WL_SetToWla(pAdaptor, uBuf+2+sizeof(TxHeader), DataPtr, templen);
                DataPtr+=templen;
                WL_GetFromWla(pAdaptor, uBuf, &uNext, sizeof(uNext));
                uBuf=uNext;
            }
            else{
                Signal.Size=packet_length+24+4-2;
                DataPtr+=2;
                packet_length-=2;
                if (packet_length>(254-sizeof(TxHeader)+6)){
                    templen=(254-sizeof(TxHeader)+6);
                    packet_length-=templen;
                }
                else{
                    templen=packet_length;
                    packet_length=0;
                }
                WL_SetToWla(pAdaptor, uBuf+2+sizeof(TxHeader)-6, DataPtr, templen);
                DataPtr+=templen;
                WL_GetFromWla(pAdaptor, uBuf, &uNext, sizeof(uNext));
                uBuf=uNext;
            }
        }
        else{
            packet_length+=12;
            Signal.Size=packet_length+24+4;
            DataPtr-=12;
            if (packet_length>(254-sizeof(TxHeader)+6)){
                templen=(254-sizeof(TxHeader)+6);
                packet_length-=templen;
            }
            else{
                templen=packet_length;
                packet_length=0;
            }
            WL_SetToWla(pAdaptor, uBuf+2+sizeof(TxHeader)-6, DataPtr, templen);
            DataPtr+=templen;
            WL_GetFromWla(pAdaptor, uBuf, &uNext, sizeof(uNext));
            uBuf=uNext;
        }
        while (packet_length >0){
            if (packet_length>254){
                templen=254;
                packet_length-=254;
            }
            else{
                templen=packet_length;
                packet_length=0;
            }
            WL_SetToWla(pAdaptor, uBuf+2, DataPtr, templen);
            DataPtr+=templen;
            WL_GetFromWla(pAdaptor, uBuf, &uNext, sizeof(uNext));
            uBuf=uNext;
        }
        WL_SetToWla(pAdaptor, uSig, &Signal, sizeof(MdReq));
        WL_ESBQRequest(pAdaptor,&uSig);
    }
    WHEN_WLAPI_DEBUG(pAdaptor->ulDbgStat &= ~WL_STAT_TX_PKT);
    return TRUE;    
}


/*
 * Receive data from Receive Queue.
 *
 * pBuf = address of host 
 * uSize = size of buffer.
 *
 * Scenario:
 *
 * Return: 0
 */
WORD WL_Receive(WL_Adaptor *pAdaptor,BYTE *pBuf, WORD uSize)
{
    WORD    NextAddress,NextAddress1;
    BYTE    *pData;
    pData=pBuf+12;
    uSize-=12;
    WL_GetFromWla(pAdaptor, pAdaptor->StartSegment+2, &NextAddress, sizeof(NextAddress));
    if (pAdaptor->LLCType==1){
        if (pAdaptor->EtherType==EtherII){
            if (uSize>BlockSize-sizeof(RxHeader)){
                WL_GetFromWla(pAdaptor, pAdaptor->StartSegment+sizeof(RxHeader), pData, BlockSize-sizeof(RxHeader));
                uSize=uSize-(BlockSize-sizeof(RxHeader));
                pData=pData+(BlockSize-sizeof(RxHeader));
            }
            else{
                WL_GetFromWla(pAdaptor, pAdaptor->StartSegment+sizeof(RxHeader), pData, uSize);
                uSize=0;
            }
        }
        else{
            uSize-=2;
            *pData=(uSize>>8) & 0xff;
            *(pData+1)=uSize & 0xff;
            pData+=2;
            if (uSize>BlockSize-sizeof(RxHeader)+6){
                WL_GetFromWla(pAdaptor, pAdaptor->StartSegment+sizeof(RxHeader)-6, pData, BlockSize-sizeof(RxHeader)+6);
                uSize=uSize-(BlockSize-sizeof(RxHeader))+6;
                pData=pData+(BlockSize-sizeof(RxHeader)+6);
            }
            else{
                WL_GetFromWla(pAdaptor, pAdaptor->StartSegment+sizeof(RxHeader)-6, pData, uSize);
                uSize=0;
            }
        }
    }
    else{
            if (uSize>BlockSize-sizeof(RxHeader)-6){
                WL_GetFromWla(pAdaptor, pAdaptor->StartSegment+sizeof(RxHeader)+6, pData, BlockSize-sizeof(RxHeader)-6);
                uSize=uSize-(BlockSize-sizeof(RxHeader))-6;
                pData=pData+(BlockSize-sizeof(RxHeader)-6);
            }
            else{
                WL_GetFromWla(pAdaptor, pAdaptor->StartSegment+sizeof(RxHeader)+6, pData, uSize);
                uSize=0;
            }

    }
    while (uSize>0){
        if (uSize>BlockSize-5){
           WL_GetFromWla(pAdaptor, NextAddress+5, pData, BlockSize-5);
           uSize=uSize-(BlockSize-5);
           pData=pData+(BlockSize-5);
           WL_GetFromWla(pAdaptor, NextAddress+2, &NextAddress1, sizeof(NextAddress1));
           NextAddress=NextAddress1;
        }
        else{
           WL_GetFromWla(pAdaptor, NextAddress+5, pData, uSize);
           uSize=0;
        }
    }
    return 0;
}


BOOL WL_ESBQRequestTest(WL_Adaptor *pAdaptor)
{
    BYTE cTmp;
    WL_GetFromWla(pAdaptor, pAdaptor->uESBQReqHead+3, &cTmp, sizeof(cTmp));
    if ((cTmp & 0x80) == 0) return FALSE;
    return TRUE;
}
void WL_ESBQRequest(WL_Adaptor *pAdaptor, WORD *ptr)
{
    WORD cTmp=0x0;
    WL_SetToWla(pAdaptor, pAdaptor->uESBQReqHead, ptr, 2);
    WL_SetToWla(pAdaptor, pAdaptor->uESBQReqHead+2, &cTmp, sizeof(cTmp));
    pAdaptor->uESBQReqHead+=4;
    if (pAdaptor->uESBQReqHead>=pAdaptor->uESBQReqEnd)
        pAdaptor->uESBQReqHead=pAdaptor->uESBQReqStart;
    return;
}
void WL_ESBQRequestFree(WL_Adaptor *pAdaptor)
{
    BYTE cTmp;
    WORD Address;
    if (pAdaptor->uESBQReqHead != pAdaptor->uESBQReqTail){
        WL_GetFromWla(pAdaptor, pAdaptor->uESBQReqTail+3, &cTmp, sizeof(cTmp));
        if ((cTmp & 0x80) !=0){
            WL_GetFromWla(pAdaptor, pAdaptor->uESBQReqTail, &Address, sizeof(Address));
            WL_FreeTxBuffer(pAdaptor,Address);
            pAdaptor->uESBQReqTail+=4;
            if (pAdaptor->uESBQReqTail>=pAdaptor->uESBQReqEnd)
                pAdaptor->uESBQReqTail=pAdaptor->uESBQReqStart;
        }
    }
    return;
}
BOOL WL_ESBQConfirm(WL_Adaptor *pAdaptor)
{
    BYTE cTmp;
    WL_GetFromWla(pAdaptor, pAdaptor->uESBQConfirm+3, &cTmp, sizeof(cTmp));
    if ((cTmp & 0x80) == 0) return FALSE;
    return TRUE;
}
void WL_ESBQConfirmDone(WL_Adaptor *pAdaptor)
{
    BYTE cTmp=0x0;
    WL_SetToWla(pAdaptor, pAdaptor->uESBQConfirm+3, &cTmp, sizeof(cTmp));
    pAdaptor->uESBQConfirm+=4;
    if (pAdaptor->uESBQConfirm>=pAdaptor->uESBQConfirmEnd)
        pAdaptor->uESBQConfirm=pAdaptor->uESBQConfirmStart;
    return;
}
BOOL MgmtStart(WL_Adaptor *pAdaptor){
    struct t_StartReq Signal;
    WORD uPtr;
    Signal.NextBlock=0;
    Signal.SignalID=StartRequest;
    memcpy((char *) Signal.SSID,(char *) pAdaptor->ESSID,34);
    memcpy((char *) pAdaptor->KeepESSID,(char *) pAdaptor->ESSID,34);
    Signal.BSSType=Independent;
    Signal.BeaconPeriod=400;
    Signal.DTIMPeriod=1;
    Signal.PHYpset[0]=3;
    Signal.PHYpset[1]=1;
    Signal.PHYpset[2]=pAdaptor->Channel;
    Signal.CapablilityInfo=0x02;
    Signal.BSSBasicRateSet[0]=0x01;
    Signal.BSSBasicRateSet[1]=0x02;
    Signal.BSSBasicRateSet[2]=0x82;
    Signal.BSSBasicRateSet[3]=0x84;
    Signal.OperationalRateSet[0]=0x01;
    Signal.OperationalRateSet[1]=0x02;
    Signal.OperationalRateSet[2]=0x82;
    Signal.OperationalRateSet[3]=0x84;
    Signal.IBSSpset[0]=6;
    Signal.IBSSpset[1]=2;
    Signal.IBSSpset[2]=10;
    Signal.IBSSpset[3]=0;
    if (WL_ESBQRequestTest(pAdaptor)){
        uPtr= WL_GetTxBuffer(pAdaptor, sizeof(Signal));
        if (uPtr !=0){
            WL_SetToWla(pAdaptor, uPtr, &Signal, sizeof(Signal));
            WL_ESBQRequest(pAdaptor,&uPtr);
            return TRUE;
        }
    }
    return FALSE;
}
BOOL MgmtScan(WL_Adaptor *pAdaptor,WORD ChannelTime){
    struct t_ScanReq Signal;
    WORD uPtr;

    Signal.NextBlock=0;
    Signal.SignalID=ScanRequest;
    Signal.SSID[0]=0;
    Signal.SSID[1]=0;
    Signal.ScanType=ActiveScan;
    Signal.ProbeDelay=0x10;
    Signal.MinChannelTime=ChannelTime;
    Signal.MaxChannelTime=ChannelTime;
    if (pAdaptor->NetworkType == Infrastructure){
        Signal.BSSType=Infrastructure;
    }
    else{
        Signal.BSSType=Independent;
    }
    pAdaptor->BSSCount=0;
    pAdaptor->JoinStaBSS=0;
    if (WL_ESBQRequestTest(pAdaptor)){
        uPtr= WL_GetTxBuffer(pAdaptor, sizeof(Signal));
        if (uPtr !=0){
            WL_SetToWla(pAdaptor, uPtr, &Signal, sizeof(Signal));
            WL_ESBQRequest(pAdaptor,&uPtr);
            return TRUE;
        }
    }
    return FALSE;
}
BOOL MgmtJoin(WL_Adaptor *pAdaptor,WORD stas){
    struct t_JoinReq Signal;
    WORD uPtr;

    Signal.NextBlock=0;
    Signal.SignalID=JoinRequest;
    Signal.Timeout=10;
    memcpy((char *) &(Signal.BeaconPeriod),(char *) &(pAdaptor->BSSset[stas].BeaconPeriod),72);
    pAdaptor->CapablilityInfo=Signal.CapablilityInfo;
    pAdaptor->Channel=Signal.PHYpset[2];
    if (WL_ESBQRequestTest(pAdaptor)){
        uPtr= WL_GetTxBuffer(pAdaptor, sizeof(Signal));
        if (uPtr !=0){
            WL_SetToWla(pAdaptor, uPtr, &Signal, sizeof(Signal));
            WL_ESBQRequest(pAdaptor,&uPtr);
            return TRUE;
        }
    }
    return FALSE;
}
BOOL MgmtAuthetication(WL_Adaptor *pAdaptor){
    struct t_AutheReq Signal;
    WORD uPtr;

    Signal.NextBlock=0;
    Signal.SignalID=AutheRequest;
    Signal.Type=Open_System;
    Signal.Timeout=1000;
    memcpy((char *) &(Signal.MacAddress),(char *) &(pAdaptor->BSSID),6);
    if (WL_ESBQRequestTest(pAdaptor)){
        uPtr= WL_GetTxBuffer(pAdaptor, sizeof(Signal));
        if (uPtr !=0){
            WL_SetToWla(pAdaptor, uPtr, &Signal, sizeof(Signal));
            WL_ESBQRequest(pAdaptor,&uPtr);
            return TRUE;
        }
    }
    return FALSE;
}
BOOL MgmtAssociation(WL_Adaptor *pAdaptor){
    struct t_AssocReq Signal;
    WORD uPtr;

    Signal.NextBlock=0;
    Signal.SignalID=AssocRequest;
    Signal.Timeout=1000;
    Signal.ListenInterval=5;
    Signal.CapInfo=pAdaptor->CapablilityInfo;
    memcpy((char *) &(Signal.MacAddress),(char *) &(pAdaptor->BSSID),6);
    if (WL_ESBQRequestTest(pAdaptor)){
        uPtr= WL_GetTxBuffer(pAdaptor, sizeof(Signal));
        if (uPtr !=0){
            WL_SetToWla(pAdaptor, uPtr, &Signal, sizeof(Signal));
            WL_ESBQRequest(pAdaptor,&uPtr);
            return TRUE;
        }
    }
    return FALSE;
}
BOOL MgmtResync(WL_Adaptor *pAdaptor){
    struct t_ResyncReq Signal;
    WORD uPtr;

    Signal.NextBlock=0;
    Signal.SignalID=ResyncRequest;
    if (WL_ESBQRequestTest(pAdaptor)){
        uPtr= WL_GetTxBuffer(pAdaptor, sizeof(Signal));
        if (uPtr !=0){
            WL_SetToWla(pAdaptor, uPtr, &Signal, sizeof(Signal));
            WL_ESBQRequest(pAdaptor,&uPtr);
            return TRUE;
        }
    }
    return FALSE;
}
void MgmtScanConfirm(WL_Adaptor *pAdaptor,WORD Address){
    WORD i,j;
    BOOL matchflag;
    struct t_ScanConf Signal;
    matchflag=FALSE;
    WL_GetFromWla(pAdaptor, Address, &Signal, sizeof(Signal));
    if (Signal.Status == SUCCESS){
        if (pAdaptor->DriverState == ScanRequest){
            for (i=0;i<pAdaptor->BSSCount;i++){
                if (memcmp((char *) pAdaptor->BSSset[i].BSSID,(char *) Signal.BSSID,6)==0)
                    break;
            }
            if ((i == pAdaptor->BSSCount) && (i<20) ){
                memcpy((char *) &(pAdaptor->BSSset[i].BeaconPeriod),(char *) &(Signal.BeaconPeriod),73);
                pAdaptor->BSSCount++;
            }
        }
        else if ( ((pAdaptor->NetworkType == Infrastructure) && ((Signal.CapablilityInfo & 0x01) != 0))
            || ((pAdaptor->NetworkType == Independent) && ((Signal.CapablilityInfo & 0x02) != 0))
            || (pAdaptor->NetworkType == 2) ){
                if (pAdaptor->ESSID[1]==0) matchflag=TRUE;
        else if ((pAdaptor->ESSID[1]==3) && (strncmp((char *) &(pAdaptor->ESSID[2]),"ANY", 3) == 0)){
            matchflag=TRUE;
        }
                else if (pAdaptor->ESSID[1] != Signal.SSID[1]) matchflag=FALSE;
                else if (memcmp((char *) &(pAdaptor->ESSID[2]),(char *) &(Signal.SSID[2]),pAdaptor->ESSID[1]))
                    matchflag=FALSE;
                else matchflag=TRUE;
                if (matchflag){
                    for (i=0;i<pAdaptor->BSSCount;i++){
                        if (memcmp((char *) pAdaptor->BSSset[i].BSSID,(char *) Signal.BSSID,6) ==0){
                            matchflag=FALSE;
                            break;
                        }
                    }
                }
                if (matchflag && (i<20)){
                    memcpy((char *) &(pAdaptor->BSSset[i].BeaconPeriod),(char *) &(Signal.BeaconPeriod),73);
                    pAdaptor->BSSCount++;
                }
        }
    }
    else if ((Signal.Status == TIMEOUT) && (pAdaptor->DriverState != ScanRequest)){
        pAdaptor->JoinStaBSS=0;
        for (j=pAdaptor->JoinStaBSS;j<pAdaptor->BSSCount;j++){
            for (i=0;i<3000;i++){
                if (MgmtJoin(pAdaptor,j)) break;
            }
            if (i != 3000) break;
        }
        pAdaptor->JoinStaBSS=j;
        if (pAdaptor->JoinStaBSS == pAdaptor->BSSCount){
            if (pAdaptor->NetworkType == Infrastructure){
                for (i=0;i<3000;i++){
                    if (MgmtScan(pAdaptor,100)) break;
                }
            }
            else{
                pAdaptor->AdHocTimes++;
                if (pAdaptor->AdHocTimes>8){
                    for (i=0;i<3000;i++){
                        if (MgmtStart(pAdaptor)) break;
                    }
                }
                else{
                    for (i=0;i<3000;i++){
                        if (MgmtScan(pAdaptor,100)) break;
                    }
                }
            }

        }

    }

}
void MgmtJoinConfirm(WL_Adaptor *pAdaptor,WORD Address){
    WORD i,j;
    struct t_JoinConf Signal;

    WL_GetFromWla(pAdaptor, Address, &Signal, sizeof(Signal));
    if (Signal.Status == SUCCESS){
        if (pAdaptor->NetworkType == Infrastructure){
            if (pAdaptor->JoinStaBSS < pAdaptor->BSSCount){
                i=pAdaptor->JoinStaBSS;
                memcpy((char *) &(pAdaptor->BSSID),(char *) &(pAdaptor->BSSset[i].BSSID),6);
                pAdaptor->Channel=pAdaptor->BSSset[i].PHYpset[2];
                memcpy((char *) pAdaptor->KeepESSID,(char *) pAdaptor->BSSset[i].SSID,34);
                for (i=0;i<3000;i++){
                    if (MgmtAuthetication(pAdaptor)) break;
                }
            }
        }
        else{
                pAdaptor->CardStart=TRUE;
                i=pAdaptor->JoinStaBSS;
                memcpy((char *) &(pAdaptor->BSSID),(char *) &(pAdaptor->BSSset[i].BSSID),6);
                pAdaptor->Channel=pAdaptor->BSSset[i].PHYpset[2];
                memcpy((char *) pAdaptor->KeepESSID,(char *) pAdaptor->BSSset[i].SSID,34);
                printk("Wireless LAN On-Line  Join:");
                printk(" %2X %2X %2X %2X %2X %2X\n",pAdaptor->BSSID.b0,pAdaptor->BSSID.b1,pAdaptor->BSSID.b2,pAdaptor->BSSID.b3,pAdaptor->BSSID.b4,pAdaptor->BSSID.b5);
        }
    }
    else{
        pAdaptor->JoinStaBSS++;
        for (j=pAdaptor->JoinStaBSS;j<pAdaptor->BSSCount;j++){
            for (i=0;i<3000;i++){
                if (MgmtJoin(pAdaptor,j)) break;
            }
            if (i != 3000) break;
        }
        pAdaptor->JoinStaBSS=j;
        if (pAdaptor->JoinStaBSS == pAdaptor->BSSCount){
            if (pAdaptor->NetworkType == Infrastructure){
                for (i=0;i<3000;i++){
                    if (MgmtScan(pAdaptor,100)) break;
                }
            }
            else{
                pAdaptor->AdHocTimes++;
                if (pAdaptor->AdHocTimes>8){
                    for (i=0;i<3000;i++){
                        if (MgmtStart(pAdaptor)) break;
                    }
                }
                else{
                    for (i=0;i<3000;i++){
                        if (MgmtScan(pAdaptor,100)) break;
                    }
                }
            }

        }

    }
}
void MgmtResyncConfirm(WL_Adaptor *pAdaptor){
    WORD i,j;
    pAdaptor->JoinStaBSS++;
    for (j=pAdaptor->JoinStaBSS;j<pAdaptor->BSSCount;j++){
        for (i=0;i<3000;i++){
            if (MgmtJoin(pAdaptor,j)) break;
        }
        if (i != 3000) break;
    }
    pAdaptor->JoinStaBSS=j;
    if (pAdaptor->JoinStaBSS >= pAdaptor->BSSCount){
        if (pAdaptor->NetworkType == Infrastructure){
            for (i=0;i<3000;i++){
                if (MgmtScan(pAdaptor,100)) break;
            }
        }
        else{
            pAdaptor->AdHocTimes++;
            if (pAdaptor->AdHocTimes>8){
                for (i=0;i<3000;i++){
                    if (MgmtStart(pAdaptor)) break;
                }
            }
            else{
                for (i=0;i<3000;i++){
                    if (MgmtScan(pAdaptor,100)) break;
                }
            }
        }

    }

}
